// Copyright 2014 The Flutter Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

import 'dart:async';

import 'package:dartweb/src/base/time.dart';
import 'package:dwds/dwds.dart';
import 'package:meta/meta.dart';

import '../application_package.dart';
import '../base/async_guard.dart';
import '../base/common.dart';
import '../base/file_system.dart';
import '../base/io.dart';
import '../base/logger.dart';
import '../base/terminal.dart';
import '../base/utils.dart';
import '../build_info.dart';
import '../dart/pub.dart';
import '../debug_option.dart';
import '../globals.dart' as globals;
import '../project.dart';
import '../resident_runner.dart';
import '../web/compile.dart';
import '../web/web_runner.dart';
import 'devfs_web.dart';

class WebApplicationPackage extends ApplicationPackage {
  WebApplicationPackage(this.flutterProject) : super(id: flutterProject.manifest.appName);

  final FlutterProject flutterProject;

  @override
  String get name => flutterProject.manifest.appName;

  /// The location of the web source assets.
  Directory get webSourcePath => flutterProject.directory.childDirectory('web');
}


/// Injectable factory to create a [ResidentWebRunner].
class DwdsWebRunnerFactory extends WebRunnerFactory {
  @override
  ResidentRunner createWebRunner({
    String target,
    String output,
    @required FlutterProject flutterProject,
    @required DebuggingOptions debuggingOptions,
  }) {
    return _ResidentWebRunner(
      target: target,
      output: output,
      flutterProject: flutterProject,
      debuggingOptions: debuggingOptions,
    );
  }
}

const String kExitMessage =  'Failed to establish connection with the application '
  'instance in Chrome.\nThis can happen if the websocket connection used by the '
  'web tooling is unable to correctly establish a connection, for example due to a firewall.';

/// A hot-runner which handles browser specific delegation.
abstract class ResidentWebRunner extends ResidentRunner {
  ResidentWebRunner({
    String target,
    String output,
    @required this.flutterProject,
    @required DebuggingOptions debuggingOptions,
  }) : super(
          target: target ?? globals.fs.path.join('lib', 'main.dart'),
          output: output,
          debuggingOptions: debuggingOptions,
        );

  final FlutterProject flutterProject;
  DateTime firstBuildTime;

  // Used with the new compiler to generate a bootstrap file containing plugins
  // and platform initialization.
  Directory _generatedEntrypointDirectory;

  // Only the debug builds of the web support the service protocol.


  @override
  bool get supportsWriteSkSL => false;



  @override
  void printHelp({bool details = true}) {
    if (details) {
      return printHelpDetails();
    }
    const String fire = '🔥';
    const String rawMessage =
        '  To hot restart changes while running, press "r". '
        'To hot restart (and refresh the browser), press "R".';
    final String message = globals.terminal.color(
      fire + globals.terminal.bolden(rawMessage),
      TerminalColor.red,
    );
    globals.printStatus(
        "Warning: Flutter's support for web development is not stable yet and hasn't");
    globals.printStatus('been thoroughly tested in production environments.');
    globals.printStatus('For more information see https://flutter.dev/web');
    globals.printStatus('');
    globals.printStatus(message);
    const String quitMessage = 'To quit, press "q".';
    globals.printStatus('For a more detailed help message, press "h". $quitMessage');
  }

}

class _ResidentWebRunner extends ResidentWebRunner {
  _ResidentWebRunner({
    String target,
    String output,
    @required FlutterProject flutterProject,
    @required DebuggingOptions debuggingOptions,
  }) : super(
          flutterProject: flutterProject,
          target: target ?? globals.fs.path.join('lib', 'main.dart'),
          output: output,
          debuggingOptions: debuggingOptions,
        );

  @override
  Future<int> run({
    Completer<DebugConnectionInfo> connectionInfoCompleter,
    Completer<void> appStartedCompleter,
    String route,
  }) async {
    firstBuildTime = DateTime.now();
    final ApplicationPackage package = await ApplicationPackageFactory.instance.getPackageForPlatform(
      TargetPlatform.web_javascript,
      applicationBinary: null,
    );
    if (package == null) {
      globals.printStatus('This application is not configured to build on the web.');
      globals.printStatus('To add web support to a project, run `flutter create .`.');
    }
    if (!globals.fs.isFileSync(mainPath)) {
      String message = 'Tried to run $mainPath, but that file does not exist.';
      if (target == null) {
        message +=
            '\nConsider using the -t option to specify the Dart file to start.';
      }
      globals.printError(message);
      return 1;
    }

    try {
      return await asyncGuard(() async {
        DateTime packagesStartTime = systemClock.now();
        print('get packages....');
        await pub.get(
          context: PubContext.pubGet
        );

        DateTime packagesEndTime = systemClock.now();
        print('pub get took ${getElapsedAsMilliseconds(packagesEndTime.difference(packagesStartTime))}.');

          await buildWeb(
            flutterProject,
            target,
            output,
            debuggingOptions.buildInfo,
            debuggingOptions.initializePlatform,
            false,
            debuggingOptions.buildInfo.dartExperiments,
          );
        return 0;
      });
    } on WebSocketException {
      throwToolExit(kExitMessage);
    } on ChromeDebugException {
      throwToolExit(kExitMessage);
    } on AppConnectionException {
      throwToolExit(kExitMessage);
    } on SocketException {
      throwToolExit(kExitMessage);
    }
    return 0;
  }

  @override
  Future<OperationResult> restart({
    bool fullRestart = false,
    bool pause = false,
    String reason,
    bool benchmarkMode = false,
  }) async {
    final Stopwatch timer = Stopwatch()..start();
    final Status status = globals.logger.startProgress(
      'Performing hot restart...',
      timeout: supportsServiceProtocol
          ? timeoutConfiguration.fastOperation
          : timeoutConfiguration.slowOperation,
      progressId: 'hot.restart',
    );

    try {
      await buildWeb(
        flutterProject,
        target,
        output,
        debuggingOptions.buildInfo,
        debuggingOptions.initializePlatform,
        false,
        debuggingOptions.buildInfo.dartExperiments,
      );
    } on ToolExit {
      return OperationResult(1, 'Failed to recompile application.');
    }
    final String elapsed = getElapsedAsMilliseconds(timer.elapsed);
    globals.printStatus('Restarted application in $elapsed.');
    return OperationResult.ok;
  }

  @override
  bool get supportsCanvasKit => supportsServiceProtocol;

}
