// Copyright 2014 The Flutter Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

import 'dart:async';
import 'package:dartweb/src/debug_option.dart';

import '../base/common.dart';
import '../build_info.dart';
import '../cache.dart';
import '../globals.dart' as globals;
import '../project.dart';
import '../resident_runner.dart';
import '../runner/flutter_command.dart';
import '../web/web_runner.dart';

abstract class RunCommandBase extends FlutterCommand {
  // Used by run and drive commands.
  RunCommandBase({ bool verboseHelp = false }) {
    addBuildModeFlags(defaultToRelease: false, verboseHelp: verboseHelp);
    usesTargetOption();
    usesOutputOption();
    usesPubOption();
  }
}

class RunCommand extends RunCommandBase {
  RunCommand({ bool verboseHelp = false }) : super(verboseHelp: verboseHelp);

  @override
  final String name = 'build';

  @override
  final String description = 'Run your Flutter app on an attached device.';


  @override
  Future<String> get usagePath async {
    final String command = await super.usagePath;
    return command;
  }


  @override
  bool get shouldRunPub {
    // If we are running with a prebuilt application, do not run pub.
    if (runningWithPrebuiltApplication) {
      return false;
    }

    return super.shouldRunPub;
  }


  bool get runningWithPrebuiltApplication => false;


  @override
  Future<void> validateCommand() async {
    // When running with a prebuilt application, no command validation is
    // necessary.
    if (!runningWithPrebuiltApplication) {
      await super.validateCommand();
    }
  }

  DebuggingOptions _createDebuggingOptions() {
    final BuildInfo buildInfo = getBuildInfo();
    return DebuggingOptions.disabled(
      buildInfo,
      hostname: '',
      port: '',
      webUseSseForDebugProxy: false,
      webEnableExposeUrl: false,
      webRunHeadless: false,
    );
  }

  @override
  Future<FlutterCommandResult> runCommand() async {
    Cache.releaseLockEarly();


    globals.terminal.usesTerminalUi = true;

    final FlutterProject flutterProject = FlutterProject.current();
    // Only support "web mode" with a single web device due to resident runner
    // refactoring required otherwise.

    ResidentRunner runner;

    runner = webRunnerFactory.createWebRunner(
      target: targetFile,
      output: outputFile,
      flutterProject: flutterProject,
      debuggingOptions: _createDebuggingOptions(),
    );

    DateTime appStartedTime;
    // Sync completer so the completing agent attaching to the resident doesn't
    // need to know about analytics.
    //
    // Do not add more operations to the future.
    final Completer<void> appStartedTimeRecorder = Completer<void>.sync();
    // This callback can't throw.

    final int result = await runner.run(
      appStartedCompleter: appStartedTimeRecorder,
    );
    if (result != 0) {
      throwToolExit(null, exitCode: result);
    }
    return FlutterCommandResult(
      ExitStatus.success,
      timingLabelParts: <String>[
        'cold',
        getModeName(getBuildMode()),
        null,
      ],
      endTimeOverride: appStartedTime,
    );
  }
}
